@@ -366,6 +366,18 @@ def project_M(M, M_constraint):
         raise NotImplementedError
     return result
 
+def project_M_(M, M_constraint):
+    result = M.clone()
+    if M_constraint == 'simplex':
+        result = project2simplex_(result, dim=0, zero_threshold=1e-5)
+    elif M_constraint == 'unit sphere':
+        result = M.div(torch.linalg.norm(result, ord=2, dim=0, keepdim=True))
+    elif M_constraint == 'nonneg unit sphere':
+        result = M.clip(1e-10).div(torch.linalg.norm(result, ord=2, dim=0, keepdim=True))
+    else:
+        raise NotImplementedError
+    return result
+
 def project2simplex(y, dim: int = 0, zero_threshold: float = 1e-10) -> torch.Tensor:
     """Projects a matrix such that the columns (or rows) lie on the unit simplex.
 
@@ -398,7 +410,6 @@ def project2simplex(y, dim: int = 0, zero_threshold: float = 1e-10) -> torch.Ten
         newton_update = objective_value / derivative
         mu -= newton_update
 
-        assert not torch.isnan(mu).any()
         previous_derivative = derivative
     assert (derivative == previous_derivative).all()
     
@@ -408,9 +419,47 @@ def project2simplex(y, dim: int = 0, zero_threshold: float = 1e-10) -> torch.Ten
     assert not torch.isnan(y).any(), (mu, derivative)
 
     assert y.sum(dim=dim).sub_(1).abs_().max() < 1e-3, y.sum(dim=dim).sub_(1).abs_().max()
-    
+   
     return y
 
+def project2simplex_(y, dim: int = 0, zero_threshold: float = 1e-10) -> torch.Tensor:
+    """(In-place) Projects a matrix such that the columns (or rows) lie on the unit simplex.
+
+    See https://math.stackexchange.com/questions/2402504/orthogonal-projection-onto-the-unit-simplex
+    for a reference.
+
+    The goal is to find a scalar mu such that || (y-mu)_+ ||_1 = 1
+
+    Currently uses Newton's method to optimize || y - mu ||^2
+
+    TODO: try implementing it this way instead: https://arxiv.org/pdf/1101.6081.pdf
+
+    Args:
+        y: list of vectors to be projected to unit simplex
+        dim: dimension along which to project
+        zero_threshold: threshold to treat as zero for numerical stability purposes
+    """
+    y_copy = y.clone()
+    num_components = y.shape[dim]
+
+    y_copy.sub_(y_copy.sum(dim=dim, keepdim=True).sub_(1), alpha=1/num_components)
+    mu = y_copy.max(dim=dim, keepdim=True)[0].div_(2)
+    derivative_prev, derivative = None, None
+    for _ in range(num_components):
+        difference = y_copy.sub(mu)
+        objective_value = difference.clip_(min=zero_threshold).sum(dim, keepdim=True).sub_(1)
+        derivative = difference.gt_(zero_threshold).sum(dim, keepdim=True)
+
+        if derivative_prev is not None and (derivative == derivative_prev).all():
+            break
+
+        mu.addcdiv_(objective_value, derivative)
+        derivative_prev = derivative
+
+    y_copy.sub_(mu).clip_(min=zero_threshold)
+    assert y_copy.sum(dim=dim).sub_(1).abs_().max() < 1e-4, y_copy.sum(dim=dim).sub_(1).abs_().max()
+    return y_copy
+
 class IndependentSet:
     """Iterator class that yields a list of batch_size independent nodes from a spatial graph.
 
@@ -423,11 +472,12 @@ class IndependentSet:
         batch_size: number of nodes to draw independently every iteration
     """
 
-    def __init__(self, adjacency_list, batch_size=50):
+    def __init__(self, adjacency_list, device, batch_size=50):
         self.N = len(adjacency_list)
         self.adjacency_list = adjacency_list
         self.batch_size = batch_size
         self.indices_remaining = None
+        self.device = device
 
     def __iter__(self):
         """Return iterator over nodes in graph.
@@ -446,18 +496,22 @@ class IndependentSet:
         if len(self.indices_remaining) == 0:
             raise StopIteration
 
-        valid_indices = []
-        excluded_indices = set()
-        effective_batch_size = min(self.batch_size, len(self.indices_remaining))
-        candidate_indices = np.random.choice(list(self.indices_remaining),
-            size=effective_batch_size,
-            replace=False,
-        )
-        for index in candidate_indices:
-            if index not in excluded_indices:
-                valid_indices.append(index)
-                excluded_indices |= set(self.adjacency_list[index])
-
+        valid_indices = sample_graph_iid(self.adjacency_list, self.indices_remaining, self.batch_size)
         self.indices_remaining -= set(valid_indices)
 
-        return valid_indices
+        return torch.tensor(valid_indices, device=self.device, dtype=torch.long)
+
+def sample_graph_iid(adjacency_list, indices_remaining, sample_size):
+    valid_indices = []
+    excluded_indices = set()
+    effective_batch_size = min(sample_size, len(indices_remaining))
+    candidate_indices = np.random.choice(list(indices_remaining),
+        size=effective_batch_size,
+        replace=False,
+    )
+    for index in candidate_indices:
+        if index not in excluded_indices:
+            valid_indices.append(index)
+            excluded_indices |= set(adjacency_list[index])
+
+    return valid_indices
